-- Background Music Manager

log("  ╠ Audio - BGM Manager")

audio_bgm = {
  fade = 1,
  fade_to = 1,
  fade_time = 2,
  
  randomized = false,
  countdown = 0,
  
  playlist = { }
}

function audio_bgm:init()
  log("Initializing BGM audio...")
  
  self.songs = { }
  
  -- Get all songs in the "audio/bgm/" directory
  
  local list = love.filesystem.getDirectoryItems("audio/bgm/")
  
  -- Prepare metadata for the songs so they can loop properly
  
  local total_songs = 0
  for i, song in ipairs(list) do
    local ext = string.sub(song, -4, -1)
    if ext == ".wav" or ext == ".ogg" then
      local new_song = {
        title = string.sub(song, 0, -5),
        loop = "audio/bgm/" .. song,
        tag = { }
      }
      
      new_song.artist = string.match(new_song.title, "([^-]+)%s%-%s")
      new_song.display_title = string.match(new_song.title, "[^-]+%s%-%s(.+)")
      
      if string.sub(song, -8, -5) == "LOOP" then -- If the song is a LOOP, add its LOOPSTART
        new_song.title = string.sub(song, 0, -10)
        new_song.display_title = string.sub(new_song.display_title, 0, -6)
        new_song.loopstart = "audio/bgm/" .. string.sub(song, 0, -9) .. "LOOPSTART" .. string.sub(song, -4)
      end
      
      if string.sub(song, -13, -5) ~= "LOOPSTART" then
        total_songs = total_songs + 1
        self.songs[new_song.title] = new_song
      end
    end
  end
  
  -- Import metadata
  
  self:importMetadata()
  
  -- Create human-readable, alphabetized list
  
  self.list = table.ofKeys(self.songs)
  table.sort(self.list)
  
  log("Loaded " .. total_songs .. " songs!")
  
  local untagged = 0
  
  for i, s in pairs(self.songs) do
    if #table.ofKeys(s.tag) == 0 then
      warn(s.title .. " is untagged!")
      untagged = untagged + 1
    end
  end
  
  if untagged > 0 then
    warn(untagged .. " songs out of " .. #self.list .. " are untagged!")
  end
end

function audio_bgm:update(dt)
  if self.fade_to > self.fade then
    self.fade = math.min(self.fade_to, self.fade + (dt / self.fade_time))
  elseif self.fade_to < self.fade then
    self.fade = math.max(self.fade_to, self.fade - (dt / self.fade_time))
  end
  
  if self.current_song then
    self.current_source:setVolume((self.current_song.vol or 1) * self.fade * settings.volume_bgm * settings.volume)
    
    if self.current_source_start then
      self.current_source_start:setVolume((self.current_song.vol or 1) * self.fade * settings.volume_bgm * settings.volume)
    end
    
    if self.fade == 0 and self.fade_to == 0 then
      self:stop()
    end
  end
  
  if self.current_source_start then
    if (not self.current_source_start:isPlaying()) then
      self.current_source:play()
    end
  end
  
  if self.countdown > 0 then
    self.countdown = math.max(0, self.countdown - dt)
    
    if self.countdown == 0 then
      if self.randomized then
        self:queue("random")
      end
      self.fade_to = 0
    end
  end
  
  if self.current_song == nil and self.queued then
    if self.queued ~= "random" and self.songs[self.queued].tag.no_fade_in then
      self.fade = 1
      self.fade_to = 1
    elseif self.fade_to == 0 then
      self.fade_to = 1
    end
    
    self:play(self.queued)
    
    self.queued = nil
  end
end

function audio_bgm:play(song)
  self:stop()
  
  if song == "random" then
    if self.playlist and #self.playlist > 0 then
      self:play(table.random(self.playlist))
    else
      local next_song = table.random(table.ofKeys(self.songs))
      
      while self.songs[next_song].tag.no_shuffle do
        next_song = table.random(table.ofKeys(self.songs))
      end
      
      self:play(next_song)
    end
  elseif self.songs[song] then
    self.current_song = self.songs[song]
    
    self.current_source = love.audio.newSource(self.current_song.loop, "static")
    self.current_source:setVolume((self.current_song.vol or 1) * self.fade * settings.volume_bgm * ((system.glitching and .5) or 1) * settings.volume)
    self.current_source:setLooping(true)
    
    if self.current_song.loopstart then
      self.current_source_start = love.audio.newSource(self.current_song.loopstart, "static")
      self.current_source_start:setVolume((self.current_song.vol or 1) * self.fade * settings.volume_bgm * ((system.glitching and .5) or 1) * settings.volume)
      self.current_source_start:play()
    else
      self.current_source:play()
    end
    
    if self.randomized then
      if audio_bgm:currentSong().tag.short_shuffle then
        self.countdown = self:currentLength(true) * (1 + (random(16, 48) / 100))
      else
        self.countdown = self:currentLength(true) * (2 + random())
      end
    end
  end
end

function audio_bgm:stop()
  if self.current_source then
    self.current_source:stop()
    
    if self.current_source_start then
      self.current_source_start:stop()
    end
    
    self.current_source = nil
    self.current_source_start = nil
    self.current_song = nil
  end
end

function audio_bgm:fadeIn()
  self.fade_to = 1
end

function audio_bgm:fadeOut()
  self.fade_to = 0
end

function audio_bgm:fadeTo(target)
  self.fade_to = target
end

function audio_bgm:queue(song)
  self.queued = song
end

function audio_bgm:randomize(setting)
  if setting == nil then
    audio_bgm.randomized = not audio_bgm.randomized
  else
    audio_bgm.randomized = setting
  end
  
  if audio_bgm.randomized and (not self.current_song) then
    self:play("random")
  elseif (not audio_bgm.randomized) then
    self.countdown = 0
  else
    if audio_bgm:currentSong().tag.short_shuffle then
      self.countdown = self:currentLength(true, true) * (1 + (random(16, 48) / 100))
    else
      self.countdown = self:currentLength(true, true) * (2 + random())
    end
  end
end

function audio_bgm:currentTitle(full)
  if self.current_song then
    if not full then
      return self.current_song.display_title
    else
      return self.current_song.title
    end
  else
    return "none"
  end
end

function audio_bgm:currentSong()
  if self.current_song then
    return self.current_song
  else
    return {title = "none", display_title = "Nothing", artist = "Nobody", loop = "none"}
  end
end

function audio_bgm:currentLoop()
  if self.current_song then
    if self.current_source_start and self.current_source_start:isPlaying() then
      return self.current_song.loopstart
    else
      return self.current_song.loop
    end
  else
    return "none"
  end
end

function audio_bgm:currentTime(raw)
  if self.current_song then
    local time = self.current_source:tell()
    
    if raw then
      return time
    else
      local mins = math.floor(time / 60)
      local secs = time % 60
      
      return string.format("%02d", mins) .. ":" .. string.format("%02d", secs)
    end
  else
    if raw then
      return 0
    else
      return "00:00"
    end
  end
end

function audio_bgm:currentLoopstartTime()
  if self.current_source_start then
    return self.current_source_start:tell()
  else
    return 0
  end
end

function audio_bgm:currentLength(raw, full)
  if self.current_song then
    local time = self.current_source:getDuration()
    
    if raw then
      if full and self.current_source_start then
        return time + self.current_source_start:getDuration()
      else
        return time
      end
    else
      local mins = math.floor(time / 60)
      local secs = time % 60
      
      return string.format("%02d", mins) .. ":" .. string.format("%02d", secs)
    end
  else
    if raw then
      return 0
    else
      return "00:00"
    end
  end
end

function audio_bgm:currentLoopstartLength()
  if self.current_source_start then
    return self.current_source_start:getDuration()
  else
    return 0
  end
end

function audio_bgm:exportMetadata()
  notify("Exporting BGM metadata...")
  
  --local success, error_msg = love.filesystem.write("bgm_meta.lua", serialize(audio_bgm.songs))
  
  --pixFile(serialize(audio_bgm.songs), false, "bgm_meta.png")
  
  local bgm_metadata_file = io.open("bgm_meta.lua", "w")
  bgm_metadata_file:write(serialize(audio_bgm.songs))
  
  local success = true
  
  if success then
    notify("BGM metadata exported successfully!")
  else
    err(error_msg)
  end
end

function audio_bgm:importMetadata()
  log("Importing BGM metadata...")
  
  if love.filesystem.getInfo("bgm_meta.lua") then
    local chunk, error_msg = love.filesystem.load("bgm_meta.lua")
    
    --local chunk, error_msg = load(pixFile("bgm_meta.png", true))
    
    if chunk then
      local import = chunk()
      
      for k, s in pairs(self.songs) do
        if import[k] then
          s.vol = import[k].vol
          s.tag = import[k].tag
          
          if s.tag.no_shuffle then
            s.tag.no_shuffle = nil
          end
        end
      end
      
      for k, s in pairs(import) do
        if (not self.songs[k]) then
          if (not love.filesystem.getInfo("audio/bgm/" .. s.loop)) then
            --warn("Metadata loaded for missing song file: \"" .. k .. "\"")
          end
          
          --self.songs[k] = s
        end
      end
      
      log("BGM metadata imported successfully!")
    else
      err(error_msg)
    end
  else
    warn("No \"bgm_meta\" file found!")
  end
end

function audio_bgm:clearPlaylist()
  self.playlist = { }
end

function audio_bgm:hasPlaylist()
  return #self.playlist > 0
end

function audio_bgm:isOnPlaylist(song)
  return table.has(self.playlist, song)
end

function audio_bgm:addToPlaylist(song)
  if self.songs[song] then
    table.insert(self.playlist, song)
  end
end

function audio_bgm:removeFromPlaylist(song)
  local has, index = table.has(self.playlist, song)
  if self.songs[song] and has then
    table.remove(self.playlist, index)
  end
end