-- Number Dial

local spinner_str = {
  blank = " ",
  numbers = "0\n1\n2\n3\n4\n5\n6\n7\n8\n9\n0",
  pos_neg = "+\n \n-",
  currency = "$",
  percent = "%",
  countdown = "T",
}

dial = { }
dial.__index = dial

setmetatable(dial, {
    __call = function(base, ...)
      local self = setmetatable({ }, base)
      self:new(...)
      return self
    end
  })

function dial:new(opt)
  self.value = opt.value or 0
  self.max = opt.max
  self.min = opt.min
  self.spinners = { }
  self.independent_spinners = opt.independent_spinners
  self.neutral_zeroes = opt.neutral_zeroes
  
  if opt.mod_spinner then
    self:newModSpinner(opt.mod_spinner)
  end
  
  if opt.currency_spinner then
    self:newModSpinner("currency")
  end
  
  if opt.percent_spinner then
    self:newModSpinner("percent")
  end
  
  for i = (opt.spinners or 1), 1, -1 do
    self:newSpinner(#self.spinners + 1, i)
  end
  
  self.w = (#self.spinners * 18) + 2
end

function dial:update(dt)
  for i, s in ipairs(self.spinners) do
    s:update(dt)
  end
end

function dial:draw(x, y)
  if x == nil then
    x = 0
  end
  if y == nil then
    y = 0
  end
  
  local x_adj = 0
  
  for i, s in ipairs(self.spinners) do
    setColor("near_black")
    love.graphics.rectangle("fill", x + x_adj, y, 20, 36)
    for i = 4, 32 do
      local p = ((32 - i + 4) / 32)
      love.graphics.setColor(HSLtoRGB((s.h or .12), (s.s or .48), .88 - (p * .24)))
      love.graphics.rectangle("fill", x + x_adj + 2, y + 2, 16, 32 * p)
    end
    s:canvasDraw()
    love.graphics.draw(s.canvas, x + x_adj, y + 2)
    
    x_adj = x_adj + 18
  end
end

function dial:getWidth()
  return self.w or 0
end

function dial:newSpinner(pos, val)
  local ns = {
    parent = self,
    canvas = love.graphics.newCanvas(20, 32),
    place = val,
    shown_value = math.floor(self.value / 10 ^ (val - 1)),
    
    update = function(self, dt)
      local t = math.floor(math.abs(self.parent.value) / 10 ^ (self.place - 1))
      
      if self.parent.independent_spinners then
        while math.abs(self.shown_value - t) > 5 do
          if self.shown_value < t then
            self.shown_value = self.shown_value + 10
          else
            self.shown_value = self.shown_value - 10
          end
        end
      end
      
      self.shown_value = lerp(self.shown_value, t, 4 * math.max(.08, math.abs(self.shown_value - t)))
      
      if self.parent.neutral_zeroes and self.shown_value == 0 then
        self.s = 0
      else
        self.s = nil
      end
    end,
    
    canvasDraw = function(self, x, y, x_adj)
      local old_canvas = love.graphics.getCanvas()
      love.graphics.setCanvas(self.canvas)
        love.graphics.clear()
        setColor("white")
        love.graphics.setFont(fonts.kubasta)
        love.graphics.printo(spinner_str.numbers, 3, -(36 * (self.shown_value % 10)))
      love.graphics.setCanvas(old_canvas)
    end
  }
  
  table.insert(self.spinners, ns)
end

function dial:newModSpinner(kind)
  local ns = {
    parent = self,
    canvas = love.graphics.newCanvas(20, 32),
    place = val,
    target_value = 0,
    shown_value = 0,
    str = kind,
    
    update = function(self, dt)
      if self.assess then
        self:assess()
      end
      
      self.shown_value = lerp(self.shown_value, self.target_value, 4 * math.max(.08, math.abs(self.shown_value - self.target_value)))
    end,
    
    canvasDraw = function(self, x, y, x_adj)
      local old_canvas = love.graphics.getCanvas()
      love.graphics.setCanvas(self.canvas)
        love.graphics.clear()
        setColor("white")
        love.graphics.setFont(fonts.kubasta)
        love.graphics.printo(spinner_str[self.str], 3, -(36 * (self.shown_value % 10)))
      love.graphics.setCanvas(old_canvas)
    end
  }
  
  if kind == "neg" then
    ns.str = "pos_neg"
    ns.target_value = 1
    ns.shown_value = 1
    ns.h = 0
    ns.assess = function(self)
      if self.parent.value < 0 then
        self.s = .64
        self.target_value = 2
      else
        self.s = 0
        self.target_value = 1
      end
    end
  elseif kind == "pos_neg" then
    ns.assess = function(self)
      if self.parent.value < 0 then
        self.h = 0
        self.s = .64
        self.target_value = 2
      elseif self.parent.value > 0 then
        self.h = .32
        self.s = .64
        self.target_value = 0
      else
        self.s = 0
        self.target_value = 1
      end
    end
  elseif kind == "currency" or kind == "percent" then
    ns.s = 0
  elseif kind == "countdown" then
    ns.str = "pos_neg"
    ns.target_value = 2
    ns.shown_value = 2
    ns.s = 0
    self:newModSpinner("countdown2")
  elseif kind == "countdown2" then
    ns.str = "countdown"
    ns.s = 0
  end
  
  table.insert(self.spinners, ns)
end

function dial:set(x)
  self.value = x
end

function dial:mod(x, min, max)
  self.value = self.value + x
  
  if min or self.min then
    self.value = math.max((self.min or min), self.value)
  end
  
  if max or self.max then
    self.value = math.min((self.max or max), self.value)
  end
end