-- GUI - Element -> Window

window = { }
window.__index = window

setmetatable(window, {
    __index = gui_element,
    __call = function(base, ...)
      local self = setmetatable({ }, base)
      self:new(...)
      return self
    end
  })

function window:new(opt)
  if not opt then
    opt = { }
  end
  
  gui_element.new(self, opt)
  
  self.tx = self.x
  self.ty = self.y
  self.tw = math.max(gui_skin_quad_w, self.w)
  self.th = math.max(gui_skin_quad_h, self.h)
  
  self.x = self.x + (self.tw / 2)
  self.y = self.y + (self.th / 2)
  self.w = 0
  self.h = 0
  
  self.state = "HIDDEN"
  self.move_speed = opt.move_speed or 16
  self.was_moving = false
  self.fx = {
    step = .02,
    speed = 1.28
  }
  self.arrow = opt.arrow
  self.instant = opt.instant
  self.skin = opt.skin or "basic"
  self.colorless = opt.colorless
  local r, g, b = HSLtoRGB(settings.window.h, settings.window.s, settings.window.l)
  self.color = {r, g, b}
  self.alpha = 1
  self.canvas = love.graphics.newCanvas(self.tw - 24, self.th - 24)
  self.off_canvas = opt.off_canvas
  
  self.canvasDraw = opt.canvasDraw
  
  self.onMoveFinish = opt.onMoveFinish
  
  return self
end

function window:update(dt)
  gui_element.update(self, dt)
  
  local r, g, b = HSLtoRGB(settings.window.h, settings.window.s, settings.window.l)
  self.color = {r, g, b}
  
  local next_state
  
  if self.state == "OPEN" then
    if not self.instant then
      audio_sfx:play("open3_mod")
      self.fx.step = .02
    else
      audio_sfx:play("msg")
      self.fx.step = 1
    end
    self.state = "OPENING"
  elseif self.state == "OPENING" then
    self.fx.step = self.fx.step + (dt * self.fx.speed)
    
    if self.fx.step >= .64 then
      self.fx.step = 1
    end
    
    local tp = easeOutElastic(self.fx.step)
    
    self.x = self.tx + (self.tw / 2) - ((self.tw / 2) * tp)
    self.y = self.ty + (self.th / 2) - ((self.th / 2) * tp)
    self.w = self.tw * tp
    self.h = self.th * tp
    
    if self.fx.step == 1 then
      next_state = "IDLE"
    end
  elseif self.state == "CLOSE" then
    if not self.instant then
      audio_sfx:play("close3")
      self.fx.step = .24
    else
      audio_sfx:play("msg")
      self.fx.step = 0
    end
    self.state = "CLOSING"
  elseif self.state == "CLOSING" then
    self.fx.step = self.fx.step - (dt * self.fx.speed)
    
    if self.fx.step <= .02 then
      self.fx.step = .02
    end
    
    local tp = easeOutElastic(self.fx.step)
    
    self.x = self.tx + (self.tw / 2) - ((self.tw / 2) * tp)
    self.y = self.ty + (self.th / 2) - ((self.th / 2) * tp)
    self.w = self.tw * tp
    self.h = self.th * tp
    
    if self.fx.step == .02 then
      self:moveToBottom()
      next_state = "HIDDEN"
    end
  elseif self.state == "IDLE" then
    if self.x ~= self.tx then
      self.x = lerp(self.x, self.tx, self.move_speed * math.max(1, math.abs(self.x - self.tx)))
      --self.x = lerp(self.x, self.tx, 4 * canvasWidth())
      self.was_moving = true
    end
    
    if self.y ~= self.ty then
      self.y = lerp(self.y, self.ty, self.move_speed * math.max(1, math.abs(self.y - self.ty)))
      --self.y = lerp(self.y, self.ty, 4 * canvasHeight())
      self.was_moving = true
    end
    
    if self.was_moving and self.x == self.tx and self.y == self.ty then
      self.was_moving = false
      
      if self.onMoveFinish then
        self:onMoveFinish()
      end
    end
  elseif self.state == "HIDDEN" then
    self.x = self.tx + (self.tw / 2)
    self.y = self.ty + (self.th / 2)
    
    if self.delete_on_hide then
      self.delete = true
    end
  end
  
  if next_state ~= nil then
    self.state = next_state
  end
end

function window:draw()
  -- Draw to window canvas
  
  if (not self.off_canvas) and self.canvasDraw then
    local old_canvas = love.graphics.getCanvas()
    love.graphics.setCanvas(self.canvas)
      love.graphics.clear()
      self:canvasDraw()
    love.graphics.setCanvas(old_canvas)
  end
  
  if self.state ~= "HIDDEN" then
    if self.skin ~= "none" then
      -- Draw window
      
      if self.colorless then
        setColor("white", self.alpha)
      else
        setColor(self.color, self.alpha)
      end
      
      local scale_w = (self.w - (gui_skin_quad_w * 2)) / gui_skin_quad_w
      local scale_h = (self.h - (gui_skin_quad_h * 2)) / gui_skin_quad_h
      
      love.graphics.draw(gui_skins[self.skin], gui_skin_quad.center.center, self.x, self.y, 0, self.w / gui_skin_quad_w, self.h / gui_skin_quad_h)
      love.graphics.draw(gui_skins[self.skin], gui_skin_quad.top.left, self.x, self.y)
      love.graphics.draw(gui_skins[self.skin], gui_skin_quad.top.center, self.x + gui_skin_quad_w, self.y, 0, scale_w, 1)
      love.graphics.draw(gui_skins[self.skin], gui_skin_quad.center.left, self.x, self.y + gui_skin_quad_h, 0, 1, scale_h)
      love.graphics.draw(gui_skins[self.skin], gui_skin_quad.center.right, self.x + self.w - gui_skin_quad_w, self.y + gui_skin_quad_h, 0, 1, scale_h)
      love.graphics.draw(gui_skins[self.skin], gui_skin_quad.bottom.left, self.x, self.y + self.h - gui_skin_quad_h)
      love.graphics.draw(gui_skins[self.skin], gui_skin_quad.bottom.center, self.x + gui_skin_quad_w, self.y + self.h - gui_skin_quad_h, 0, scale_w, 1)
      love.graphics.draw(gui_skins[self.skin], gui_skin_quad.bottom.right, self.x + self.w - gui_skin_quad_w, self.y + self.h - gui_skin_quad_h)
      love.graphics.draw(gui_skins[self.skin], gui_skin_quad.top.right, self.x + self.w - gui_skin_quad_w, self.y)
    end
    
    if self.arrow and self.state == "IDLE" then
      if cursor_blink < .8 then
        if assess(self.arrow.up) then
          love.graphics.drawCentered(gui_arrow, self.x - .5 + (self.w / 2), self.y - 1)
        end
        if assess(self.arrow.down) then
          love.graphics.drawCentered(gui_arrow, self.x - .5 + (self.w / 2), self.y + self.h + 1, math.pi)
        end
        if assess(self.arrow.right) then
          love.graphics.drawCentered(gui_arrow, self.x + self.w + 1, self.y - .5 + (self.h / 2), math.rad(90))
        end
        if assess(self.arrow.left) then
          love.graphics.drawCentered(gui_arrow, self.x - 1, self.y - .5 + (self.h / 2), math.rad(270))
        end
      else
        if assess(self.arrow.up) then
          love.graphics.drawCentered(gui_arrow, self.x - .5 + (self.w / 2), self.y + 2)
        end
        if assess(self.arrow.down) then
          love.graphics.drawCentered(gui_arrow, self.x - .5 + (self.w / 2), self.y + self.h - 2, math.pi)
        end
        if assess(self.arrow.right) then
          love.graphics.drawCentered(gui_arrow, self.x + self.w - 2, self.y - .5 + (self.h / 2), math.rad(90))
        end
        if assess(self.arrow.left) then
          love.graphics.drawCentered(gui_arrow, self.x + 2, self.y - .5 + (self.h / 2), math.rad(270))
        end
      end
    end
    
    -- Draw window canvas
    
    setColor("white")
    love.graphics.draw(self.canvas, self.x + 12, self.y + 12, 0, (self.w - 24) / self.canvas:getWidth(), (self.h - 24) / self.canvas:getHeight())
    
    if self.off_canvas then
      love.graphics.push()
        love.graphics.translate(self.x + 12, self.y + 12)
        love.graphics.scale((self.w - 24) / self.canvas:getWidth(), (self.h - 24) / self.canvas:getHeight())
        self:canvasDraw()
      love.graphics.pop()
    end
  end
end

function window:centerX()
  self.tx = (game_canvas:getWidth() / 2) - (self.tw / 2)
  self.x = self.x + (self.tw / 2)
  
  return self
end

function window:centerY()
  self.ty = (game_canvas:getHeight() / 2) - (self.th / 2)
  self.y = self.y + (self.th / 2)
  
  return self
end

function window:move(x, y, speed)
  self.tx = x or self.tx
  self.ty = y or self.ty
  self.move_speed = speed or self.move_speed
  
  return self
end

function window:open()
  if self.state == "HIDDEN" then
    self.state = "OPEN"
  end
  
  return self
end

function window:forceOpen()
  self.state = "OPEN"
  
  return self
end

function window:close(delete)
  if self.state == "IDLE" then
    self.state = "CLOSE"
  end
  
  if delete then
    self.delete_on_hide = true
  end
  
  return self
end

function window:forceClose(delete)
  self.state = "CLOSE"
  
  if delete then
    self.delete_on_hide = true
  end
  
  return self
end