-- GUI - Glyph

local rainbow_speed = .32

glyph = { }
glyph.__index = glyph

setmetatable(glyph, {
    __call = function(base, ...)
      local self = setmetatable({ }, base)
      self:new(...)
      return self
    end
  })

function glyph:new(opt)
  self.char = opt.char
  self.icon = opt.icon
  self.x = opt.x
  self.y = opt.y
  self.w = opt.w
  self.h = opt.h
  
  self.new_line = opt.new_line
  
  self.shake_x = 0
  self.shake_y = 0
  self.jump_y = 0
  self.jump_t = 0
  self.jump_p = 0
  self.rainbow_h = opt.rainbow_h
  self.flash_t = 0
  
  if self.char then
    if not opt.font then
      opt.font = game_font --love.graphics.getFont()
    end
    
    self.w = opt.font:getWidth(self.char)
    self.h = opt.font:getHeight(self.char)
  end
  
  self.fx = opt.fx or { }
  self.hold = opt.hold
  self.action = opt.action
  
  return self
end

function glyph:update(dt)
  if self.fx.shake then
    self.shake_x = random(-1, 1)
    self.shake_y = random(-1, 1)
  end
  
  if self.fx.jump and self.fx.jump ~= -1 then
    self.jump_t = self.jump_t - dt
    
    if self.jump_t <= 0 then
      --self.jump_t = self.fx.jump
      
      self.jump_p = math.min(2, self.jump_p + (dt * 2))
      
      self.jump_y = 32 * easeOutSine(self.jump_p)
      
      if self.jump_p == 2 then
        self.jump_p = 0
        self.jump_t = self.fx.jump
      end
    end
  end
  
  if self.fx.rainbow then
    self.rainbow_h = (self.rainbow_h + (dt * rainbow_speed)) % 1
  end
  
  if self.fx.ephemeral and self.fx.ephemeral > 0 then
    self.fx.ephemeral = math.max(0, self.fx.ephemeral - dt)
  end
  
  if self.fx.flash then
    self.flash_t = (self.flash_t + dt) % 2
  end
end

function glyph:draw(offset_x, offset_y)
  local _, _, _, a = love.graphics.getColor()
  if self.fx.rainbow then
    love.graphics.setColor(HSLtoRGB(self.rainbow_h, 1, .5, a))
  end
  
  if self.fx.ephemeral then
    if self.fx.ephemeral == 0 then
      return
    end
    
    local r, g, b, a = love.graphics.getColor()
    
    love.graphics.setColor(r, g, b, math.min(1, self.fx.ephemeral))
  elseif self.fx.flash then
    local r, g, b, a = love.graphics.getColor()
    
    love.graphics.setColor(r, g, b, (math.sin((math.pi / 2) * (4 * self.flash_t + 1)) + 1) / 2)
  end
  
  if self.char then
    love.graphics.printo(self.char, self.x + offset_x + self.shake_x, self.y + offset_y + self.shake_y - self.jump_y)
  elseif self.icon then
    drawIcon(self.icon, self.x + offset_x + self.shake_x, self.y + offset_y + self.shake_y - self.jump_y)
  end
end

-- GUI - Text

text = { }
text.__index = text

setmetatable(text, {
    __call = function(base, ...)
      local self = setmetatable({ }, base)
      self:new(...)
      return self
    end
  })

function text:new(opt)
  self.text = opt.text
  self.x = opt.x or 0
  self.y = opt.y or 0
  self.w = opt.w or 1280
  self.line_h = opt.line_h or ((opt.font and (opt.font:getHeight() + 2)) or (game_font:getHeight() + 2)) --(love.graphics.getFont():getHeight() + 2)
  self.max_w = 0
  self.max_h = 0
  self.glyphs = { }
  self.on_glyph = 0
  self.tbg = opt.time or .016
  self.time = self.tbg
  self.hold = 0
  self.instant = opt.instant
  self.alpha = opt.alpha
  
  self.font = opt.font or game_font --love.graphics.getFont()
  
  self.fx = {
    shake = false,
    jump = -1,
    rainbow = false
  }
  self.next_rainbow_h = 0
  
  self:parse()
  
  return self
end

function text:update(dt)
  if self.instant then
    local jump_offset = 0
    
    while self.on_glyph < #self.glyphs and self.instant do
      self.on_glyph = math.min(#self.glyphs, self.on_glyph + 1)
      local g = self.glyphs[self.on_glyph]
      
      if g.fx and g.fx.jump ~= -1 then
        jump_offset = jump_offset + self.tbg
        g.jump_t = g.jump_t + jump_offset
      end
      
      if self.glyphs[self.on_glyph].action then
        self.glyphs[self.on_glyph].action()
      end
    end
  end
  
  if self.hold == 0 then
    self.time = self.time - dt
    
    if self.time <= 0 then
      self.on_glyph = math.min(#self.glyphs, self.on_glyph + 1)
      self.time = self.tbg
      
      if self.glyphs[self.on_glyph].hold then
        self.hold = self.glyphs[self.on_glyph].hold
      end
      
      if self.glyphs[self.on_glyph].action then
        self.glyphs[self.on_glyph].action()
      end
    end
  elseif self.hold > 0 then
    self.hold = self.hold - dt
    
    if self.hold <= 0 then
      self.hold = 0
    end
  end
  
  if self.on_glyph > 0 then
    for i = 1, self.on_glyph do
      local g = self.glyphs[i]
      
      if g.update then
        g:update(dt)
      end
    end
  end
end

function text:draw()
  local old_canvas = love.graphics.getCanvas()
  
  if self.glitch_canvas then
    love.graphics.setCanvas(self.glitch_canvas)
      love.graphics.clear()
    love.graphics.setCanvas(old_canvas)
  end
  
  setColor("white", (self.alpha or 1))
  love.graphics.setFont(self.font)
  
  if self.on_glyph > 0 then
    for i = 1, self.on_glyph do
      local g = self.glyphs[i]
      
      if g.fx and g.fx.glitch then
        love.graphics.setCanvas(self.glitch_canvas)
          g:draw(0, 0)
        love.graphics.setCanvas(old_canvas)
      elseif g.draw then
        g:draw(self.x, self.y)
      else
        setColor(g, (self.alpha or 1))
      end
    end
  end
  
  if self.glitch_canvas then
    setColor("white", 1)
    local old_shader = love.graphics.getShader()
    
    love.graphics.setShader(glitch_shader_proc)
    love.graphics.draw(self.glitch_canvas, self.x, self.y)
    love.graphics.setShader(old_shader)
  end
end

function text:getWidth()
  return self.max_w + 2
end

function text:getHeight()
  --return self:prevGlyph().y + self:prevGlyph().h + 4
  return self.max_h
end

function text:getCenterX()
  return self.x + (self:getWidth() / 2)
end

function text:getCenterY()
  return self.y + (self:getHeight() / 2)
end

function text:getCenter()
  return self:getCenterX(), self:getCenterY()
end

function text:isFinished()
  return self.on_glyph == #self.glyphs and self.hold == 0
end

function text:isHolding(return_num)
  if return_num then
    return self.hold
  else
    return self.hold == -1
  end
end

function text:prevGlyph(index)
  if index == nil then
    index = #self.glyphs
  elseif index == 1 then
    return
  end
  
  repeat
    index = index - 1
  until self.glyphs[index].x ~= nil
  
  return self.glyphs[index]
end

function text:parse()
  local next_pos = {x = 0, y = 2}
  
  if type(self.text) == "string" then
    for i = 1, string.len(self.text) do
      local char = string.sub(self.text, i, i)
      local new_glyph = glyph({char = char, x = next_pos.x, y = next_pos.y, font = self.font, fx = table.deepcopy(self.fx)})
      table.insert(self.glyphs, new_glyph)
      next_pos.x = new_glyph.x + new_glyph.w
    end
  elseif type(self.text) == "table" then
    for _, val in ipairs(self.text) do
      if type(val) == "string" then
        local command = (string.sub(val, 1, 1) == "/" and string.sub(val, 2, 2) ~= "/")
        
        if command then
          local iter = string.gmatch(string.sub(val, 2, -1), "%S+")
          command = iter()
          
          if command == "nl" then
            local new_glyph = glyph({new_line = true})
            table.insert(self.glyphs, new_glyph)
          elseif command == "icon" then
            local new_glyph = glyph({icon = tonumber(iter()), x = next_pos.x, y = next_pos.y - 4, w = 32, h = 32, fx = table.deepcopy(self.fx)})
            table.insert(self.glyphs, new_glyph)
            next_pos.x = new_glyph.x + new_glyph.w + 2
          elseif command == "time" then
            table.insert(self.glyphs, glyph({action = function() self.tbg = tonumber(iter() or .016) end}))
          elseif command == "instant" then
            table.insert(self.glyphs, glyph({action = function() self.instant = tobool(iter()) end}))
          elseif command == "hold" then
            table.insert(self.glyphs, glyph({hold = tonumber(iter() or -1)}))
          elseif command == "shake" then
            self.fx.shake = tobool(iter())
          elseif command == "jump" then
            self.fx.jump = tonumber(iter())
          elseif command == "rainbow" then
            self.fx.rainbow = tobool(iter())
          elseif command == "flash" then
            self.fx.flash = tobool(iter())
          elseif command == "ephemeral" then
            local num = iter()
            if num then
              self.fx.ephemeral = tonumber(num)
            else
              self.fx.ephemeral = nil
            end
          elseif command == "glitch" then
            self.fx.glitch = tobool(iter())
            self.has_glitch = true
          end
        else
          for i = 1, string.len(val) do
            local char = string.sub(val, i, i)
            local new_glyph = glyph({char = char, x = next_pos.x, y = next_pos.y, font = self.font, fx = table.deepcopy(self.fx), rainbow_h = (self.fx.rainbow and math.abs(self.next_rainbow_h))})
            table.insert(self.glyphs, new_glyph)
            next_pos.x = new_glyph.x + new_glyph.w
            
            if self.fx.rainbow then
              self.next_rainbow_h = (self.next_rainbow_h - .05) % 1
            end
          end
        end
      elseif type(val) == "table" then
        table.insert(self.glyphs, val)
      end
    end
  end
  
  -- Assess line breaks
  
  local last_breakable = -1
  
  for i, g in ipairs(self.glyphs) do
    if g.char and (g.char == " " or g.char == "-" or g.char == "_") then
      last_breakable = i
    elseif g.new_line or (g.x and g.x + g.w > self.w and last_breakable ~= -1) then
      if g.new_line then
        last_breakable = i
      end
      
      local first = true
      
      for j = last_breakable + 1, #self.glyphs do
        local prev_glyph = self:prevGlyph(j)
        local cur_glyph = self.glyphs[j]
        
        if cur_glyph.x then
          if first then
            cur_glyph.x = 0
            first = false
          else
            cur_glyph.x = prev_glyph.x + prev_glyph.w
          end
          cur_glyph.y = cur_glyph.y + self.line_h
        end
      end
    end
  end
  
  -- Assess maximums
  
  self.max_w = 0
  self.max_h = 0
  for i, g in ipairs(self.glyphs) do
    if g.char and g.char == " " then
      -- Do nothing, space will kill the spacing
    else
      if g.x and g.x + g.w > self.max_w then
        self.max_w = g.x + g.w
      end
      
      if g.y and g.y + g.h > self.max_h then
        self.max_h = g.y + g.h
      end
    end
  end
  
  -- Create glitch canvas if necessary
  
  if self.has_glitch then
    self.glitch_canvas = love.graphics.newCanvas(self.w, self.max_h)
  end
end